# An experiment comparing PyRTL's native matrix multiply against a row-vector based approach
# inspired by the paper "Vectorization for Digital Signal Processors via Equality Saturation",
# which uses Tensilica instrinsics for achieve significant speedups.

from typing import List

import pyrtl.rtllib.matrix as Matrix
import pyrtl

def LOAD(matrix: Matrix.Matrix, start: int, count: int) -> Matrix.Matrix:
    """ Load subportion of a matrix into a row vector.

    :param Matrix matrix: matrix to load from
    :param int start: element number to start from (counting elements row-wise)
    :param int count: number of elements to take from matrix
    :return: a row vector Matrix (size 1xN)

    Given
        a = [[a0, a1, a2],
             [a3, a4, a5]]

    LOAD(a, 1, 4) returns

        [[a1, a2, a3, a4]]
    """
    flattened = matrix.flatten()
    return flattened[0, start:start+count]

def STORE(vec: Matrix.Matrix, matrix: Matrix.Matrix, start: int, count: int) -> None:
    """ Store a row vector into a matrix starting from a particular
    element index in the matrix row-wise.

    :param Matrix vec: row vector (size 1xN) to store
    :param Matrix matrix: matrix where vec will be stored; must have at least N elements
    :param int start: where to start putting elements into matrix
    :param int count: how many elements from vec to put into matrix

    The argument `matrix` will be updated as a result (no return value).

    Given
        vec = [[v0, v1, v2, v3]]
        matrix = [[m0, m1, m2],
                  [m3, m4, m5]]
    
    STORE(vec, matrix, 0, 4) will change matrix to be
        matrix = [[v0, v1, v2],
                  [v3, m4, m5]]
    """
    matrix.put(list(range(start, start+count)), vec)

def SHUF(vec: Matrix.Matrix, poss: List[int]) -> Matrix.Matrix:
    """ Shuffle the elements of a row-vector according to some given indices

    :param Matrix vec: a row vector containing elements we'll access
    :param list[int] poss: a list of column indices into vec for accessing elements
    :return Matrix: a new row vector

    Given
        v = [[v0, v1, v2, v3]]
    and
        p = [0, 1, 2, 0]

    SHUF(v, p) returns

        [[v0, v1, v2, v0]]
    """
    return Matrix.concatenate(Matrix.Matrix(1, 1, vec.bits, value=vec[0, i]) for i in poss)

def SHUF2(a: Matrix.Matrix, b: Matrix.Matrix, poss: List[int]) -> Matrix.Matrix:
    """ Helper that simply combines a and b into a single vector before calling SHUF

    :param a: a row vector containing elements we'll access
    :param b: a row vector containing elements we'll access
    :param list[int] poss: a list of column indices into vec for accessing elements
    :return: a row vector Matrix

    See SHUF for an example usage.
    """
    return SHUF(Matrix.hstack(a, b), poss)

def MUL(a: Matrix.Matrix, b: Matrix.Matrix) -> Matrix.Matrix:
    """ Does *element-wise* multiplication of two row vectors.

    :param Matrix a: a row vector Matrix (size 1xN)
    :param Matrix b: a row vector Matrix (size 1xN)
    :return: a new matrix of size 1xN

    Given
        a = [a0, a1, a2, a3]
    and
        b = [b0, b1, b2, b3]
    
    mul(a, b) returns
        [a0*b0, a1*b1, a2*b2, a3*b3]
    
    Both a and b must be vectors of the same size
    """
    return a * b

def MAC(accumulator: Matrix.Matrix, a: Matrix.Matrix, b: Matrix.Matrix) -> None:
    """ Multiply-accumulate operation.

    :param Matrix accumulator: row vector to add to
    :param Matrix a: row vector to multiply
    :param Matrix b: row vector to multiply
    
    Updates accumulator to be accum + a*b (does not return any value).
    """
    accumulator += MUL(a, b)

# Using these 'intrinsics', example from presentation
# video for ASPLOS 2021 paper "Vectorization for Digital Signal Processors via Equality Saturation".
def example_matrix_multiply(a, b, c):
    a0_4 = LOAD(a, 0, 4)
    b0_4 = LOAD(b, 0, 4)
    b4_6 = LOAD(b, 4, 2)

    ashuf1 = SHUF(a0_4, [0, 0, 0, 2])
    bshuf1 = SHUF(b0_4, [0, 1, 2, 0])
    mul1 = MUL(ashuf1, bshuf1)

    ashuf2 = SHUF(a0_4, [1, 1, 1, 3])
    bshuf2 = SHUF2(b0_4, b4_6, [3, 4, 5, 3])
    MAC(mul1, ashuf2, bshuf2)

    ashuf3 = SHUF(a0_4, [2, 2])
    bshuf3 = SHUF(b0_4, [1, 2])
    mul2 = MUL(ashuf3, bshuf3)

    ashuf4 = SHUF(a0_4, [3, 3])
    bshuf4 = SHUF(b4_6, [0, 1])
    MAC(mul2, ashuf4, bshuf4)

    STORE(mul1, c, 0, 4)
    STORE(mul2, c, 4, 2)

def print_info(name):
    ta = pyrtl.TimingAnalysis()
    print(f"Max frequency: {ta.max_freq()} MhZ")
    print(f"Max timing delay: {ta.max_length()} ps")
    print(f"Logic count: {len({l for l in pyrtl.working_block().logic if l.op not in 'wcs'})}")
    logic, _ = pyrtl.area_estimation()
    print(f"Logic area est: {logic}")
    with open(f"{name}.v", 'w') as f:
        pyrtl.output_to_verilog(f)
    with open(f"{name}.gv", 'w') as f:
        pyrtl.output_to_graphviz(f)


a_vals = [[1, 2], [3, 4]]
b_vals = [[2, 4, 6], [8, 10, 12]]


# 1) Using the special approach
print("\nSpecial approach for multiplying 2x2 and 2x3 matrices")
a_in = pyrtl.Input(4 * 4, 'a_in')
b_in = pyrtl.Input(6 * 4, 'b_in')

# Pass in wires, rather than constants (which is also acceptable),
# so that pyrtl.optimize()'s constant propagation doesn't get rid of everything,
# and so we can have a fair comparison.
a = Matrix.Matrix(2, 2, 4, value=a_in)
b = Matrix.Matrix(2, 3, 4, value=b_in)
c = Matrix.Matrix(2, 3, 32)
example_matrix_multiply(a, b, c)
output = pyrtl.Output(name='output')
output <<= c.to_wirevector()
sim = pyrtl.Simulation()
sim.step({
    'a_in': Matrix.list_to_int(a_vals, 4),
    'b_in': Matrix.list_to_int(b_vals, 4),
})
raw_matrix = Matrix.matrix_wv_to_list(sim.inspect('output'), c.rows, c.columns, c.bits)
print(raw_matrix)

pyrtl.synthesize()
print_info('custom_pre-opt')

print('')

pyrtl.optimize()
print_info('custom_post-opt')

# 2) Using Pyrtl's matrix multiply
pyrtl.reset_working_block()

print("\nGeneral PyRTL approach using __matmul__ on Matrix objects")
a00, a01, a10, a11 = pyrtl.input_list('a00/4 a01/4 a10/4 a11/4')
b00, b01, b02, b10, b11, b12 = pyrtl.input_list('b00/4 b01/4 b02/4 b10/4 b11/4 b12/4')
# Pass in wires, rather than constants (which is also acceptable),
# so that pyrtl.optimize()'s constant propagation doesn't get rid of everything,
# and so we can have a fair comparison.
# Doing this in a little different way that Example 1) above to show
# another way of passing in data to construct the matrix.
a = Matrix.Matrix(2, 2, 4, value=[[a00, a01], [a10, a11]])
b = Matrix.Matrix(2, 3, 4, value=[[b00, b01, b02], [b10, b11, b12]])
output = pyrtl.Output(name='output')
c = a @ b
output <<= c.to_wirevector()
sim = pyrtl.Simulation()
sim.step({
    'a00': a_vals[0][0],
    'a01': a_vals[0][1],
    'a10': a_vals[1][0],
    'a11': a_vals[1][1],
    'b00': b_vals[0][0],
    'b01': b_vals[0][1],
    'b02': b_vals[0][2],
    'b10': b_vals[1][0],
    'b11': b_vals[1][1],
    'b12': b_vals[1][2],
})
raw_matrix = Matrix.matrix_wv_to_list(sim.inspect('output'), c.rows, c.columns, c.bits)
print(raw_matrix)

pyrtl.synthesize()
print_info('pyrtl_pre-opt')

print('')

pyrtl.optimize()
print_info('pyrtl_post-opt')

# 3) Using Python
c = [[0, 0, 0], [0, 0, 0]]

print("\nIn regular Python, just to compare actual results")
def mm(a, b, o):
    for row in range(len(a)):
        for col in range(len(b[0])):
            s = 0
            assert len(a[0]) == len(b)
            for i in range(len(b)):
                s += a[row][i] * b[i][col]
            o[row][col] = s
mm(a_vals, b_vals, c)
print(c)

# 4) Using numpy
print("\nUsing Python's numpy, because why not?")
import numpy
a = numpy.matrix(a_vals)
b = numpy.matrix(b_vals)
c = a @ b
print(c)
